/* kill mblist, use subscribed list */  
#include "prjlibs-c/standards.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <signal.h>
#include <dirent.h>
#include <stddef.h>
#include <time.h>
#include <string.h>

#include <skalibs/stddjb.h>
#include "prjlibs-c/constants.h"
#include "prjlibs-c/case.h"
#include "prjlibs-c/warn.h"
#include "prjlibs-c/intattr.h"
#include "prjlibs-c/nsec.h"
#include "prjlibs-c/diewarn.h"
#include "prjlibs-c/types.h"
#include "prjlibs-c/vec.h"
#include "types.h"
#include "constants.h"
#include "drop_root.h"
#include "fmt0.h"
#include "get_addr.h"
#include "timeout.h"
#include "keys.h"
#include "b0.h"
#include "must.h"
#include "uil.h"
#include "input/input.h"
#include "imap/imap.h"
#include "imap-ro/imap-ro.h"

ccp PROG="posterity-imap";

static void debug(ccp a, ccp b, ccp c, ccp d, type_len e) {
  char z[e+1];
  memcpy(z, d, e);
  z[e]='\0';
  strerr_warn7x(a, ":", b, ": ", c, d, z);
}
#define debug2(a, b, c, d, e) debug(a, #b, c, d, e)
#define debug1(a, b, c, d, e) debug2(a, b, c, d, e)
#define debug(c, d, e) debug1(__FILE__, __LINE__, c, d, e)

static type_len buf_chr(bufp buf, int ch) {
  return byte_chr(buffer_RPEEK(buf), buffer_len(buf), (ch));
}

static void fill_if_empty(bufp buf, ccp src) {
  if (buffer_isempty(buf)) must_fill(buf, src);
}

static void fill_if_empty_0(void) { fill_if_empty(buffer_0, src_stdin); }

static type_bool consumec(char ch) {
  fill_if_empty_0();
  if (CASE_TO_LOWER(b0peek[0])!=CASE_TO_LOWER(ch)) return 0;
  b0seek(1);
  return 1;
}

static type_bool consumeb(ccp data, type_len len) {
  type_len off=0;
  for (;;) {
    if (b0len>=len) {
      if (case_diffb(b0peek+off, len, data)) return 0;
      b0seek(len);
      return 1;
    }
    if (case_diffb(b0peek+off, b0len, data)) return 0;
    off+=b0len;
    len-=b0len;
    data+=b0len;
    b0fillx();
  }
}

#define consumea(data) \
  (WARN_ARR(char const,data), consumeb((data), (sizeof (data))-1))

static char const capab[]="CAPABILITY IMAP4rev1 AUTH=ANONYMOUS";
/* An IMAP4rev1 server is NOT required to support obsolete IMAP4 or IMAP2bis 
   commands, though some do */

static type_uil msgs_all     =VEC_0(type_uid);
static type_uil msgs_Seen    =VEC_0(type_uid);
static type_uil msgs_Answered=VEC_0(type_uid);
static type_uil msgs_Flagged =VEC_0(type_uid);
static type_uil msgs_Deleted =VEC_0(type_uid);
static type_uil msgs_Draft   =VEC_0(type_uid);
static type_uil msgs_Recent  =VEC_0(type_uid);

typedef struct mailbox mailbox_t;
struct mailbox {
  mailbox_t* next;
  type_bool flag_subscribed:1;
  type_bool flag_exists:1;
  type_len name_len;
  char name[1];
};
#define MAILBOX_SIZE(len) (offsetof(mailbox_t, name)+(len)+1)
static mailbox_t* mb_list=NULLP, * mb_selected; 

static char buffer_space[BUFFER_INSIZE];
static type_uint64_buf pid_buf;
static ccp remote_addr="";
static str tag=VEC_0(char), path=VEC_0(char), scratch=VEC_0(char);
static type_bool flag_uid;
static struct taia mb_list_updated=TAIA_ZERO;

static void update_mb_list(void) {
  DIR* dir;
  struct stat st;
  struct taia last_mod=TAIA_ZERO, now;
  mailbox_t* new_list=NULLP;
  taia_now(&now);
  if (stat(".", &st)!=0) DIE1(stat, ".");
  tai_unix(taia_secp(&last_mod), st.st_mtime);
  taia_nano(&last_mod)=nsec_mtime(&st);
  if (taia_less(&last_mod, &mb_list_updated)) return;
  dir=opendir(".");
  if (dir==NULLP) DIE1(open, ".");
  for (;;) {
    struct dirent const* de;
    mailbox_t* mb, ** mbp;
    type_len len;
    errno=0;
    de=readdir(dir);
    if (de==NULLP) {
      if (errno==0) break;
      DIE1(read, ".");
    }
    if (de->d_name[0]=='.') continue;
    len=str_len(de->d_name);
    for (mbp=&mb_list; *mbp!=NULLP; mbp=&(*mbp)->next)
      if ((*mbp)->name_len==len && byte_equal(de->d_name, len, (*mbp)->name))
        break;
    if (*mbp!=NULLP) {
      mb=*mbp;
      *mbp=mb->next;
      mb->flag_exists=1;
      mb->next=new_list;
      new_list=mb;
    } else {
      type_len size;
      ++len;
      if (len==0) DIE_OVERFLOW();
      size=MAILBOX_SIZE(len);
      if (size<len) DIE_OVERFLOW();
      mb=(void*)alloc(size);
      if (mb==NULLP) DIE0(alloc);
      mb->flag_subscribed=0;
      mb->flag_exists=1;
      byte_copy(mb->name, len, de->d_name);
      mb->name_len=len-1;
      mb->next=new_list;
      new_list=mb;
    }
  }
  closedir(dir);
  while (mb_list!=NULLP) {
    mailbox_t* const next=mb_list->next;
    if (!mb_list->flag_subscribed) alloc_free(mb_list);
    else {
      mb_list->next=new_list;
      new_list=mb_list;
      new_list->flag_exists=0;
    }
    mb_list=next;
  }
  mb_list=new_list;
  mb_list_updated=now;
  --tai_sec(taia_secp(&mb_list_updated));
}

static type_len fill_upto_x(ccp delim, type_len dlen) {
  type_len len=0;
  for (;;) {
    len+=byte_in(b0peek+len, b0len-len, delim, dlen);
    if (len<b0len) return len;
    b0fillx();
  }
}
#define fill_upto(delim) \
  (WARN_ARR(char const, delim), fill_upto_x((delim), (sizeof (delim))-1))
#define fill_word() fill_upto(" \r\n")
#define fill_line() fill_upto("\r\n")

static type_len fill_line_limit(type_len limit) {
  type_len len=0;
  for (;;) {
    len+=byte_in(b0peek+len, b0len-len, "\r\n", 2);
    if (len<b0len || len>=limit) return len;
    b0fillx();
  }
}

static mailbox_t* find_mailbox(str const* name) {
  mailbox_t* mb;
  for (mb=mb_list; mb!=NULLP; mb=mb->next)
    if (mb->name_len==l(name) && byte_diff(sl(name), mb->name)==0)
      break;
  return mb;
}

enum state {
  state_unauthenticated=1,
  state_authenticated  =2,
  state_selected       =4,
  states_authsel=state_authenticated|state_selected,
  states_all    =state_unauthenticated|state_authenticated|state_selected,
};
static enum state state=state_unauthenticated;

typedef void command_function(void);

typedef struct {
  command_function* func;
  ccp name;
  type_len namelen;
  enum state legal_states;
  type_bool uid_ok:1;
} const type_command;
static type_command* cmd;

static str next_uid_sa=VEC_0(char);

static type_uid next_uid(ccp mb_name, type_len mb_len) { /* just use largest+1? maybe have to */ 
  static char const uid_space[sizeof (type_uid_buf)]="";
  type_len uid_offset;
  type_uid uid_opened, uid_linked;
  ino_t ino_opened;
  VEC_TRUNC(char, &next_uid_sa, 0);
  must_n(&next_uid_sa, mb_name, mb_len);
  must_a(&next_uid_sa, "/.posterity/uidlock/", 1);
  uid_offset=l(&next_uid_sa);
  must_a(&next_uid_sa, uid_space, 0);
  for (uid_opened=1;;) {
    struct stat st;
    type_uid nlink;
    fmt_uid(s(&next_uid_sa)+uid_offset, uid_opened);
    if (stat(s(&next_uid_sa), &st)!=0) {
      if (errno==ENOENT) return uid_opened;
      WARN1(stat, s(&next_uid_sa));
      return 0;
    }
    ino_opened=st.st_ino;
    nlink=st.st_nlink;
    uid_linked=uid_opened+nlink;
    if (INTATTR_NEGATIVE(nlink_t, st.st_nlink) ||
        (nlink_t)nlink!=st.st_nlink || uid_linked<nlink)
      DIE_OVERFLOW();
    fmt_uid(s(&next_uid_sa)+uid_offset, uid_linked);
    if (stat(s(&next_uid_sa), &st)!=0) {
      if (errno==ENOENT) return uid_linked;
      WARN1(stat, s(&next_uid_sa));
      return 0;
    }
    if (st.st_ino!=ino_opened) uid_opened=uid_linked; /* this is a new file */
    /* otherwise, we lost a race; try again */
  }
}

static char const sp_anon[]=" anonymous";

static type_command commands[]={
#define CMDX(name, uid_ok, states) CMD2(name, imap_readonly,   uid_ok, states)
#define CMD1(name, uid_ok, states) CMD2(name, imap_cmd_##name, uid_ok, states)
#define CMD2(name, func, uid_ok, states) \
  { func, #name, (sizeof #name)-1, states, uid_ok }
  CMD1(CAPABILITY,               0, states_all           ),
  CMD1(NOOP,                     0, states_all           ),
  CMD1(LOGOUT,                   0, states_all           ),
  CMD1(AUTHENTICATE,             0, state_unauthenticated),
  CMD1(LOGIN,                    0, state_unauthenticated),
  CMD1(SUBSCRIBE,                0, states_authsel       ),
  CMD1(UNSUBSCRIBE,              0, states_authsel       ),
  CMD1(LIST,                     0, states_authsel       ),
  CMD1(LSUB,                     0, states_authsel       ),
  CMD1(STATUS,                   0, states_authsel       ),
  CMD1(SELECT,                   0, states_authsel       ),
  CMD2(EXAMINE, imap_cmd_SELECT, 0, states_authsel       ),
  CMDX(APPEND,                   0, states_authsel       ),
  CMDX(CREATE,                   0, states_authsel       ),
  CMDX(DELETE,                   0, states_authsel       ),
  CMDX(RENAME,                   0, states_authsel       ),
  CMD1(CHECK,                    0, state_selected       ),
  CMD1(CLOSE,                    0, state_selected       ),
  CMD1(STORE,                    1, state_selected       ),
  CMD1(FETCH,                    1, state_selected       ),
  CMD1(SEARCH,                   1, state_selected       ),
  CMDX(COPY,                     1, state_selected       ),
  CMDX(EXPUNGE,                  0, state_selected       ),
#undef CMD1
#undef CMD2
#undef CMDX
};

static unsigned int const n_cmds=sizeof commands/sizeof commands[0];

int main(int argc, char** argv) {
  (void)argv;
  if (argc>1) strerr_die2x(100, "usage: tcpserver 0 143 ", PROG);

  drop_root();
  FMT0(uint64, &pid_buf, getpid());
  remote_addr=get_addr();
  {
    struct sigaction action;
    if (sigemptyset(&action.sa_mask)!=0) DIE0(sigemptyset);
    action.sa_flags=0;
    action.sa_handler=imap_handler_alrm;
    if (sigaction(SIGALRM, &action, NULLP)!=0) DIE0(sigaction);
  }
  RESPOND_NOTAG_3("OK [", capab, "] posterity");
  timeout_setup();

  for (;;) {
    type_len wordlen;
    command_function* func;

    /* read tag name */
    wordlen=fill_word();
    VEC_TRUNC(char, &tag, 0);
    must_n(&tag, b0peek, wordlen);
    must_at(&tag, "");
    b0seek(wordlen);
    timeout_reset();

    if (!consumec(' ')) { reject(); continue; }

    /* read command name */
    flag_uid=consumea("uid ");
    wordlen=fill_word();
    for (cmd=&commands[0]; cmd!=&commands[n_cmds]; ++cmd)
      if (wordlen==cmd->namelen && case_equalb(cmd->name, wordlen, b0peek))
        break;
    b0seek(wordlen);

    if (cmd==&commands[n_cmds] || (flag_uid && !cmd->uid_ok) ||
        (cmd->legal_states&state)!=state)
      { reject(); continue; }

    func=cmd->func;
    if (func!=imap_cmd_SELECT && func!=imap_cmd_CLOSE &&
        state==state_selected) {
      /* A server MUST send mailbox size updates automatically if a mailbox
         size change is observed during the processing of a command.
         Special rules exist for server notification of a client about the
         removal of messages to prevent synchronization errors; see the
         description of the EXPUNGE response for more detail. */
      /* EXISTS EXPUNGE(?) RECENT */ 
    }
    func();

    VEC_FREE(char, &path);
    VEC_FREE(char, &scratch);
  }
}

/* Other compatibility issues with IMAP2bis, the most common variant of
   the earlier protocol, are discussed in [IMAP-COMPAT]. */
/*
       Characters are 7-bit US-ASCII unless otherwise specified.  Other
       character sets are indicated using a "CHARSET", as described in
       [MIME-IMT] and defined in [CHARSET].  CHARSETs have important
       additional semantics in addition to defining character set; refer
       to these documents for more detail.

   In modified UTF-7, printable US-ASCII characters except for "&"
   represent themselves; that is, characters with octet values 0x20-0x25
   and 0x27-0x7e.  The character "&" (0x26) is represented by the two-
   octet sequence "&-".

   All other characters (octet values 0x00-0x1f, 0x7f-0xff, and all
   Unicode 16-bit octets) are represented in modified BASE64, with a
   further modification from [UTF-7] that "," is used instead of "/".
   Modified BASE64 MUST NOT be used to represent any printing US-ASCII
   character which can represent itself.

   "&" is used to shift to modified BASE64 and "-" to shift back to US-
   ASCII.  All names start in US-ASCII, and MUST end in US-ASCII (that
   is, a name that ends with a Unicode 16-bit octet MUST end with a "-
   ").      For example, here is a mailbox name which mixes English, Japanese,
      and Chinese text: ~peter/mail/&ZeVnLIqe-/&U,BTFw-
*/
/*
4.1.2 The server MAY allow the EXPUNGE of a multi-accessed mailbox,
      and on subsequent FETCH commands return FETCH responses only for
      non-expunged messages and a tagged NO.

   After receiving a tagged NO FETCH response, the client SHOULD issue a
   NOOP command so that it will be informed of any pending EXPUNGE
   responses.  The client may then either reissue the failed FETCH
   command, or by examining the EXPUNGE response from the NOOP and the
   FETCH response from the FETCH, determine that the FETCH failed
   because of pending expunges.

   Example:  (Building upon the scenario outlined in 4.1.)

   <Client #2 attempts to FETCH a mix of expunged and non-expunged
   messages.  A FETCH response is returned only for then non-expunged
   messages along with a tagged NO>

             C2: B001 FETCH 3:5 ENVELOPE
             S2: * 3 FETCH ENVELOPE . . . (ENVELOPE info returned)
             S2: B001 NO Some of the requested messages no longer exist

   <Upon receiving a tagged NO FETCH response, Client #2 issues a NOOP
   to be informed of any pending EXPUNGE responses>

             C2: B002 NOOP
             S2: * 4 EXPUNGE
             S2: * 4 EXPUNGE
             S2: * 4 EXPUNGE
             S2: * 4 EXPUNGE
             S2: * 3 EXISTS
             S2: B002 OK NOOP Completed.

   <By receiving a FETCH response for message 3, and an EXPUNGE response
   that indicates messages 4:7 have been expunged, the client does not
   need to re-issue the FETCH>

4.2.1 If the ".SILENT" suffix is used, and the STORE completed
      successfully for all the non-expunged messages, the server SHOULD
      return a tagged OK.
4.2.2. If the ".SILENT" suffix is not used, and only expunged messages
       are referenced, the server SHOULD return only a tagged NO.
4.2.3. If the ".SILENT" suffix is not used, and a mixture of expunged
       and non-expunged messages are referenced, the server MAY set the
       flags and return a FETCH response for the non-expunged messages
       along with a tagged NO.
   A server MAY simply not return a search response for messages that
   have been expunged and it has not been able to inform the client
   about.
4.4.2 The server MAY allow the COPY of messages in a multi-access
      mailbox that contains expunged messages.

   Pending EXPUNGE response(s) MUST be returned to the COPY command.
*/
