#include "prjlibs-c/standards.h"
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <stddef.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>

#include <skalibs/stddjb.h>
#include "prjlibs-c/constants.h"
#include "prjlibs-c/diewarn.h"
#include "prjlibs-c/case.h"
#include "prjlibs-c/nsec.h"
#include "prjlibs-c/vec.h"
#include "prjlibs-c/types.h"
#include "drop_root.h"
#include "fmt0.h"
#include "get_addr.h"
#include "read_line.h"
#include "respond.h"
#include "timeout.h"
#include "types.h"

ccp PROG="posterity-pop";

#define RESPOND1(r0) respond((r0), NULLP, NULLP, NULLP, NULLP, NULLP)
#define RESPOND3(r0, r1, r2)     respond((r0), (r1), (r2), NULLP, NULLP, NULLP)
#define RESPOND4(r0, r1, r2, r3) respond((r0), (r1), (r2), (r3),  NULLP, NULLP)

enum state { state_auth, state_auth_user, state_trans };
static enum state state=state_auth;
#define ASSERT_STATE(assertion) \
  do if (state!=(assertion)) { \
    RESPOND1("-ERR that command is not valid in this state"); \
    return; \
  } while (0)

typedef struct message mesg_t;
struct message {
  off_t size;
  char name[1];
};
#define MESG_SIZE(namelen) (offsetof(mesg_t, name)+namelen)

typedef struct messagep mesgp_t;
struct messagep {
  mesg_t* mesg;
  mesgp_t* next;
  time_t mtime_secs;
  unsigned int mtime_nsecs;
};

static mesgp_t* mesgp_list=NULLP;
static mesg_t** mesg_vec;
static size_t n_mesgs=0;
static VEC_T(char) path=VEC_0(char);
static size_t path_dir_len;
static type_uint64_buf pid_buf;
static type_uint64_buf total_size_buf;
static ccp remote_addr="";
static char const ok_generic[]="+OK successful";

static mesg_t* get_mesg(ccp args) {
  uint64 n;
  type_len i;
  i=uint64_scan_base(args, &n, 10);
  if (i==0 || args[i]!='\0' || n==0 || n>n_mesgs) {
    RESPOND1("-ERR no such message");
    return NULLP;
  }
  --n;
  return mesg_vec[n];
}

static void cmd_quit(char* args) {
  (void)args;
  RESPOND1("+OK quitting");
  _exit(0);
}

static void cmd_noop(char* args) {
  (void)args;
  ASSERT_STATE(state_trans);
  RESPOND1(ok_generic);
}
#define cmd_rset cmd_noop

static void cmd_user(char* args) {
  DIR* dirh;
  mesg_t** mesgv_tail;
  uint64 total_size=0;
  size_t mesgv_size;
  ASSERT_STATE(state_auth);
  if (VEC_TRUNC(char, &path, 0)!=0 ||
      VEC_PUSHS(&path, args, 1)!=0 ||
      VEC_PUSHA(char, &path, "/new/", 0)!=0)
    DIE0(alloc);
  VEC_POP1(char, &path);
  path_dir_len=VEC_LEN(char, &path);
  if (args[0]=='.' || args[str_chr(args, '/')]=='/') nobox: {
    RESPOND1("-ERR no such mailbox");
    return;
  }
  dirh=opendir(VEC_ELTS(char, &path));
  if (dirh==NULLP) {
    if (errno==ENOENT) goto nobox;
    RESPOND1("-ERR unable to open mailbox");
    DIE1(open, VEC_ELTS(char, &path));
  }
  for (;;) {
    struct stat statbuf;
    struct dirent* entry;
    mesgp_t* mesgp, ** mesgp_tail;
    mesg_t* mesg;
    size_t namelen;
    errno=0;
    entry=readdir(dirh);
    if (entry==NULLP) {
      if (errno==0) break;
      RESPOND1("-ERR unable to read mailbox");
      DIE1(read, VEC_ELTS(char, &path));
    }
    if (entry->d_name[0]=='.') continue;
    if (VEC_TRUNC(char, &path, path_dir_len)!=0 ||
        VEC_PUSHS(&path, entry->d_name, 0)!=0)
      DIE0(alloc);
    if (stat(VEC_ELTS(char, &path), &statbuf)!=0) {
      RESPOND1("-ERR unable to stat message");
      DIE1(stat, VEC_ELTS(char, &path));
    }
    total_size+=(uint64)statbuf.st_size;
    namelen=strlen(entry->d_name);
    if (++n_mesgs==0 || total_size<(uint64)statbuf.st_size ||
        entry->d_name[namelen]!='\0' || ++namelen==0) {
      RESPOND1("-ERR arithmetic overflow");
      DIE_OVERFLOW();
    }
    mesgp=malloc(sizeof *mesgp);
    mesg=malloc(MESG_SIZE(namelen));
    if (mesgp==NULLP || mesg==NULLP) {
      RESPOND1("-ERR unable to allocate memory");
      DIE0(alloc);
    }
    mesg->size=statbuf.st_size;
    memcpy(mesg->name, entry->d_name, namelen);
    mesgp->mesg=mesg;
    mesgp->mtime_secs=statbuf.st_mtime;
    mesgp->mtime_nsecs=nsec_mtime(&statbuf);
    for (mesgp_tail=&mesgp_list;; mesgp_tail=&(*mesgp_tail)->next) {
      if (*mesgp_tail==NULLP) break;
      if ((*mesgp_tail)->mtime_secs <mesgp->mtime_secs)    continue;
      if ((*mesgp_tail)->mtime_secs >mesgp->mtime_secs)    break;
      if ((*mesgp_tail)->mtime_nsecs<mesgp->mtime_nsecs)   continue;
      if ((*mesgp_tail)->mtime_nsecs>mesgp->mtime_nsecs)   break;
      if (strcmp((*mesgp_tail)->mesg->name, mesg->name)>0) break;
    }
    mesgp->next=*mesgp_tail;
    *mesgp_tail=mesgp;
  }
  closedir(dirh);
  fmt0_uint64(&total_size_buf, total_size);
  mesgv_size=n_mesgs*sizeof (mesg_t*);
  if (mesgv_size/sizeof (mesg_t*)!=n_mesgs) {
    RESPOND1("-ERR arithmetic overflow");
    DIE_OVERFLOW();
  }
  if (n_mesgs==0) mesg_vec=NULLP;
  else {
    mesg_vec=malloc(mesgv_size);
    if (mesg_vec==NULLP) DIE0(alloc);
  }
  for (mesgv_tail=mesg_vec; mesgp_list!=NULLP; ++mesgv_tail) {
    mesgp_t* next=mesgp_list->next;
    *mesgv_tail=mesgp_list->mesg;
    free(mesgp_list);
    mesgp_list=next;
  }
  state=state_auth_user;
  RESPOND1(ok_generic);
}

static void cmd_pass(char* args) {
  (void)args;
  ASSERT_STATE(state_auth_user);
  state=state_trans;
  RESPOND1(ok_generic);
}

static void cmd_stat(char* args) {
  type_uint64_buf ct_buf;
  (void)args;
  ASSERT_STATE(state_trans);
  fmt0_uint64(&ct_buf, n_mesgs);
  RESPOND4("+OK ", ct_buf, " ", total_size_buf);
}

static void list_line(ccp num_buf, mesg_t* mesg) {
  type_uint64_buf size_buf;
  fmt0_uint64(&size_buf, mesg->size);
  RESPOND3(num_buf, " ", size_buf);
}

static void uidl_line(ccp num_buf, mesg_t* mesg) {
  RESPOND3(num_buf, " ", mesg->name);
}

static void multi_line(void (*line)(ccp, mesg_t*), char* args) {
  ASSERT_STATE(state_trans);
  if (*args=='\0') {
    size_t i;
    RESPOND1(ok_generic);
    for (i=0; i!=n_mesgs; ++i) {
      type_uint64_buf num_buf;
      fmt0_uint64(&num_buf, i+1);
      line(num_buf, mesg_vec[i]);
    }
    RESPOND1(".");
  } else {
    mesg_t* const mesg=get_mesg(args);
    if (mesg!=NULLP) {
      mustwrite("+OK ");
      line(args, mesg);
    }
  }
}

static void cmd_list(char* args) { multi_line(list_line, args); }
static void cmd_uidl(char* args) { multi_line(uidl_line, args); }

static void send_mesg(mesg_t const* mesg, type_bool flag_top,
                      uint64 lines) {
  static char buffer_space[8192];
  buffer input;
  type_fd fd;
  type_bool flag_body=0;
  char last_seen[5]="xx\r\n";
  if (VEC_TRUNC(char, &path, path_dir_len)!=0 ||
      VEC_PUSHS(&path, mesg->name, 0)!=0)
    DIE0(alloc);
  fd=open(VEC_ELTS(char, &path), O_RDONLY);
  if (fd<0) {
    RESPOND1("-ERR unable to open message");
    WARN1(open, VEC_ELTS(char, &path));
    return;
  }
  strerr_warn5x(pid_buf, " ", remote_addr, " read ", VEC_ELTS(char, &path));
  RESPOND1("+OK message data follows");
  buffer_init(&input, &fd_readv, fd, buffer_space, sizeof(buffer_space));
  for (;;) {
    char ch;
    type_status nread=buffer_get(&input, &ch, 1);
    if (nread==0) break;
    if (nread<0) DIE1(read, mesg->name);
    last_seen[0]=last_seen[1];
    last_seen[1]=last_seen[2];
    last_seen[2]=last_seen[3];
    last_seen[3]=ch;
    if (strcmp(last_seen+2, "\n.")==0) mustwrite(".");
    else if (!flag_body && strcmp(last_seen, "\r\n\r\n")==0) flag_body=1;
    mustwrite(last_seen+3);
    if (flag_top && flag_body && last_seen[3]=='\n' && lines--==0) break;
  }
  fd_close(fd);
  if (strcmp(last_seen+2, "\r\n")!=0) mustwrite("\r\n");
  RESPOND1(".");
}

static void cmd_top(char* args) {
  mesg_t const* mesg;
  uint64_t lines;
  type_len i;
  ASSERT_STATE(state_trans);
  i=str_chr(args, ' ');
  if (args[i]!=' ') {
    RESPOND1("-ERR bad syntax");
    return;
  }
  ((char*)args)[i]='\0';
  mesg=get_mesg(args);
  if (mesg==NULLP) return;
  args+=i+1;
  i=uint64_scan(args, &lines);
  if (i==0 || args[i]!='\0') {
    RESPOND1("-ERR bad syntax");
    return;
  }
  send_mesg(mesg, 1, lines);
}

static void cmd_retr(char* args) {
  mesg_t* mesg;
  ASSERT_STATE(state_trans);
  mesg=get_mesg(args);
  if (mesg==NULLP) return;
  send_mesg(mesg, 0, 0);
}

static void cmd_dele(char* args) {
  (void)args;
  ASSERT_STATE(state_trans);
  RESPOND1("-ERR this is a read-only service");
}

typedef struct {
  char name[5];
  void (*fn)(char*);
} command;

#define CMD(name) { #name, cmd_##name }
static command const commands[]={
  CMD(quit),
  CMD(user),
  CMD(pass),
  CMD(noop),
  CMD(list),
  CMD(uidl),
  CMD(stat),
  CMD(top),
  CMD(retr),
  CMD(dele),
  CMD(rset),
};
#undef CMD

static size_t const n_cmds=sizeof commands/sizeof commands[0];

int main(int argc, char** argv) {
  VEC_T(char) v=VEC_0(char);

  (void)argv;
  if (argc>1) DIE_USAGE("");

  drop_root();
  fmt0_uint64(&pid_buf, getpid());
  remote_addr=get_addr();
  RESPOND1("+OK posterity");
  timeout_setup();

  for (;;) {
    char* x;
    size_t len, i;
    VEC_TRUNC(char, &v, 0);
    READ_LINE(&v);
    if (VEC_PUSHA(char, &v, "", 0)!=0) DIE0(alloc);

    timeout_reset();

    len=VEC_LEN(char, &v)-1;
    if (memchr(VEC_ELTS(char, &v), '\0', len)!=NULLP) {
      RESPOND1("-ERR bad syntax");
      continue;
    }
    for (i=0; i!=len; ++i) {
      x=VEC_ELT(char, &v, i);
      *x=CASE_TO_LOWER(*x);
    }

    x=memchr(VEC_ELTS(char, &v), ' ', len);
    if (x!=NULLP) len=x-VEC_ELTS(char, &v);
    else x=VEC_ELT(char, &v, len);

    for (i=0; i!=n_cmds; ++i)
      if (strncmp(commands[i].name, VEC_ELTS(char, &v), len)==0 &&
          len==strlen(commands[i].name))
        break;

    if (i==n_cmds) RESPOND1("-ERR unknown command");
    else commands[i].fn(x+(*x==' '));
  }
}
